#!/usr/bin/python

#   bitopret: An interpreter for the Bito programming language
#   Copyright (C) 2009  Niels Serup

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


#   You can contact Niels at <ns@metanohi.org>.
#   v0.9

import sys

# First argument does not matter
args = sys.argv[1:]
args_len = len(args)

help_arg = args and '--help' in args
# Display help
if not args or help_arg:
    print_example_usage = False
    if help_arg and args_len > 1:
        # Maybe there's a 'x' argument after '--help'
        crucial_arg_now = False
        for x in args:
            if crucial_arg_now:
                if x == 'x':
                    print_example_usage = True
                break
            elif x == '--help':
                crucial_arg_now = True
    
    print \
"""Usage: bitopret [OPTION]... [FILE]
bitopret interprets files written in the Bito programming language."""
    if not print_example_usage:
        print \
"""
  -u, --unpack   Unpack bytes into ones and zeroes before interpreting.
                 bitopret will automatically unpack bytes if the filename
                 extension of FILE is .btp, unless a -r argument is specified.
  -p, --pack     Pack every 8 bit characters into bytes, output and exit.
  -P, --print    Output Bito code and exit.
  -v, --verbose  Output interpreter actions.
  -a, --actions  Output interpreter actions. Do not output prints from code.
  -s, --silent   Do not output error messages.
  -o <filename>, --output=<filename>
                 Redirect standard out to <filename>.
  -e <filename>, --errput=<filename>
                 Redirect standard error to <filename>.
  -i <text>, --input=<text>
                 Read from <text> instead of FILE. If both <text> and FILE is
                 specified, FILE has precedence.
  -r, --read     Do not unpack bytes. Default.

  --help [x]     If no extra argument is not set, display this help and exit.
                 If 'x' is set, display example usage and exit.
  --version      Display version information and exit."""
    else:
        print \
"""
Example usage:

  $ bitopret testfile
    Will execute code found in 'testfile'.
  $ bitopret -v testfile
    Will execute code found in 'testfile' and print interpreter actions.
  $ bitopret -u packedfile
    Will unpack 'packedfile', then execute unpacked code.
  $ bitopret packedfile.btp
    Will unpack 'packedfile.btp', then execute unpacked code.
  $ bitopret -po testfile_packed testfile
    Will pack code found in 'testfile', then print it to 'testfile_packed'.
  $ bitopret -P testfile
    Will print code found in 'testfile'.
  $ bitopret -o outfile --input=01000111
    Will read 01000111 as code and redirect output to 'outfile'.
  $ bitopret -as file
    Will execute code found in 'file', but will only print interpreter actions.
  $ bitopret --errput=error_file
    Will output errors to 'error_file'."""
    sys.exit()

# Display version information
if '--version' in args:
    print \
"""bitopret 0.9
Copyright (C) 2009 Niels Serup
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law."""
    sys.exit()

from math import log, floor

UNPACK  = False # Unpack bytes into ones and zeroes before interpreting.
PACK    = False # Pack every 8 bit characters into bytes, output and exit.
PRINT   = False # Output Bito code and exit.
VERBOSE = False # Output interpreter actions.
ACTIONS = False # Output interpreter actions. Do not output prints from code.
SILENT  = False # Do not output error messages.
READ    = False # Do not unpack bytes. Default.

bito_output_filename = ''
bito_errput_filename = ''
bito_code = ''
bito_filename = ''

# Some options need info from two arguments.
special_values = []

for setting in args:
    if special_values:
        for value in special_values:
            if value == 'output':
                bito_output_filename = setting
            if value == 'errput':
                bito_errput_filename = setting
            elif value == 'input':
                bito_code = setting
            
            special_values.remove(value)
            break
        continue
           
    
    if setting == '--unpack':
        UNPACK  = True
    elif setting == '--pack':
        PACK    = True
    elif setting == '--print':
        PRINT   = True
    elif setting == '--verbose':
        VERBOSE = True
    elif setting == '--actions':
        ACTIONS = True
    elif setting == '--silent':
        SILENT  = True
    elif setting.find('--output=') == 0:
        bito_output_filename = setting[9:]
    elif setting.find('--input=') == 0:
        bito_code = setting[8:]
    elif setting == '--read':
        READ  = True
    elif setting[0] == '-':
        
        for short in setting[1:]:
            if short == 'u':
                UNPACK  = True
            elif short == 'p':
                PACK    = True
            elif short == 'P':
                PRINT   = True
            elif short == 'v':
                VERBOSE = True
            elif short == 'a':
                ACTIONS = True
            elif short == 's':
                SILENT  = True
            elif short == 'o':
                special_values.append('output')
            elif short == 'e':
                special_values.append('errput')
            elif short == 'i':
                special_values.append('input')
            elif short == 'r':
                READ = True
    else:
        bito_filename = setting

if ACTIONS:
    VERBOSE = True


# --------- #
# FUNCTIONS #
def put(txt):
    global ACTIONS
    if ACTIONS:
        return 0
    global VERBOSE
    if not VERBOSE:
        sys.stdout.write(txt)
        print '',
    else:
        print txt

def err(msg, point = -1):
    global SILENT
    if not SILENT:
        if point == -1:
            point = ''
        else:
            point = ' [' + str(point) + ']'
        
        txt = 'bitopret: ' + msg + '.' + point
        global err_to_out
        if not err_to_out:
            sys.stderr.write(txt+'\n')
        else:
            print txt

def decbin(n, l = 0):
    if not l:
        # Calculate necessary length
        l = int(floor(log(n, 2)))
    
    return ''.join([str((n >> b) & 1) for b in range(l, -1, -1)])
# FUNCTIONS #
# --------- #


# Print error messages to stdout if output file and errput file are the same
if bito_output_filename and bito_output_filename == bito_errput_filename:
    err_to_out = True
else:
    err_to_out = False

# Test for output file
if bito_output_filename:
    try:
        sys.stdout = open(bito_output_filename, 'a')
    except:
        err('cannot access ' + bito_output_filename)
        sys.exit(1)
# Test for errput file
if bito_errput_filename:
    try:
        sys.stderr = open(bito_errput_filename, 'a')
    except:
        err('cannot access ' + bito_errput_filename)
        sys.exit(1)


if not bito_filename and not bito_code:
    err('no type of input specified')
    sys.exit()

# See if there's someting in stdin
if not sys.stdin.isatty():
    stdintxt = sys.stdin.read()
else:
    stdintxt = ''
stdintxt_len = len(stdintxt)

if bito_filename:
    if bito_filename[-4:] == '.btp' and not READ:
        UNPACK = True
    
    try:
        # 'U' is 'Universal newline support'. See <http://www.python.org/dev/peps/pep-0278/>.
        bito_file = open(bito_filename, 'U')
        bito_code = bito_file.read()
        bito_code = bito_code[:-1] # Newline deletion
        bito_file.close()
        del bito_file
    except:
        err('cannot access ' + bito_filename)
        sys.exit(1)

if UNPACK:
    new_bito_code = ''
    for byte in bito_code:
        # Convert to binary in the form of xxxxxxxx
        new_bito_code += decbin(ord(byte), 7)
    bito_code = new_bito_code
    del new_bito_code

if not (READ and PRINT):
    new_bito_code = ''
    # Filter code
    for char in bito_code:
        if '01'.find(char) != -1:
            new_bito_code += char

    bito_code = new_bito_code
    del new_bito_code

# Error checking
if not bito_code:
    err('no Bito code found in ' + bito_filename)
    sys.exit(1)

bito_code_len = len(bito_code)

if not (READ and PRINT):
    if bito_code_len % 4:
        err('length of Bito code not directly dividable with four')
        sys.exit(1)


first_part_len = bito_code_len / 4
last_part_len = bito_code_len - first_part_len
first_part = bito_code[:first_part_len]
last_part = bito_code[-last_part_len:]


if PACK and not UNPACK:
    # If code is not directly dividable with 8, add an effect-less command
    if bito_code_len % 8:
        # 1 101: End loop
        bito_code = first_part + '1101' + last_part
        bito_code_len += 4
    
    packed = ''
    for i in range(0, bito_code_len, 8):
        packed += chr(int(bito_code[i:i+8], 2))
    print packed
    sys.exit()

if PRINT:
    print bito_code
    sys.exit()


del bito_code


index = 0
memlist = {'0':''}

# To allow calling last_part[-3:0], which is not possible, in the loop below,
# we add a character to last_part and instead call last_part[-4:-1].
last_part += '_'

# By default, no loop exists.
loop_start = -1

i = 0
ti = -4

while i < first_part_len:
    first = first_part[i]
    last = last_part[ti:ti+3]
    q = first + last
    
    mem = memlist[str(index)]
    
    rmem = mem
    if not rmem:
        # Replace with underscore if number is unset
        rmem = '_'
    
    try: dec = int(mem, 2)
    except: dec = -1
    
    try: ascdec = chr(dec)
    except: ascdec = ''
    
    point = ' [' + str(i) + ']'
    
    if first == '0':
        # Append xxx to current memory cell (as a string)
        xxx = last[::-1]
        memlist[str(index)] += xxx
        if VERBOSE:
            mem = memlist[str(index)]
            dec = int(mem, 2)
            
            try: asc = ' (' + repr(chr(dec)) + ')'
            except: asc = ''
            
            print 'Append ' + xxx + \
            '. Now ' + mem + ' (' + str(dec) + ')' + asc + '.' + point
    
    elif last == '000':
        # Print current number
        if VERBOSE:
            print 'Print ' + rmem + ' (' + str(dec) + ').' + point
        if not mem:
            err('current number is unset', i)
        elif not ACTIONS:
            put(str(dec))
    
    elif last == '100':
        # Print current number as ASCII
        if VERBOSE:
            print 'Print as ASCII ' + rmem + \
            ' (' + str(dec) + ') (' + repr(ascdec) + ').' + point
        if not mem:
            err('current number is unset', i)
        elif not ACTIONS:
            if ascdec:
                put(ascdec)
            else:
                err('current number is not an ASCII value', i)
    
    elif last == '010' or last == '110':
        # Increment or decrement memory cell index
        if last == '010':
            add = 1
            inorde = 'In'#crement
        else:
            add = -1
            inorde = 'De'#crement
        
        index += add
        strindex = str(index)
        if not strindex in memlist:
            memlist[strindex] = ''
        if VERBOSE:
            mem = memlist[strindex]
            
            try: dec = int(mem, 2)
            except: dec = -1
            
            try: asc = ' (' + repr(chr(dec)) + ')'
            except: asc = ''
            
            if not mem:
                numdet = 'unset'
            else:
                numdet = mem + ' (' + str(dec) + ')' + asc
            
            print inorde + 'crement index to ' + strindex +\
             '. Current number is ' + numdet + '.' + point
    
    elif last == '001':
        # Start loop
        if VERBOSE:
            print 'Start loop.' + point
        if mem:
            t = dec
        else:
            t = -1
        if loop_start == -1:
            loop_start = [i, ti, t]
    
    elif last == '101':
        # Restart or end loop
        smsg = 'End'
        eor = i
        if loop_start != -1:
            if mem:
                if loop_start[2] > 1:
                    i = loop_start[0]
                    ti = loop_start[1]
                    smsg = 'Restart'
                    eor = i
                    loop_start[2] -= 1
                else:
                    loop_start = -1
            emsg = ''
        else:
            emsg = 'non-existent '
        if VERBOSE:
            print smsg + ' ' + emsg + 'loop at point ' + str(eor) + '.' + point
    
    elif last == '011':
        # Add value of previous cell to current cell (as an integer)
        try: predec = int(memlist[str(index-1)], 2)
        except: predec = -1
        
        error = 0
        if not mem:
            error = 1
            ndec = dec
        else:
            if dec > predec:
                hdeci = index
            else:
                hdeci = index - 1
            
            ndec = dec + predec
            if ndec > -1:            
                memlist[str(index)] = decbin(ndec, len(memlist[str(hdeci)])+1)
            else:
                ndec = dec
        
        if VERBOSE:
            print 'Add value of cell ' + str(index - 1) + ' (' + str(predec) \
             + ') to cell ' + str(index) + ' (' + str(dec) + \
              '). Now ' + str(ndec) + '.' + point
        
        if error:
            err('current number is unset', i)
    
    elif last == '111':
        # Get input from stdin and save ASCII values in current and next cells
        if VERBOSE:
            print 'Write stdin to next cells, length to current. Length: ' \
            + str(stdintxt_len) + '.' + point
        if stdintxt:
            ix = index + 1
            for char in stdintxt:
                strix = str(ix)
                if not strix in memlist:
                    memlist[strix] = ''
                memlist[strix] = decbin(ord(char), 7)
                ix += 1
            memlist[str(index)] = decbin(stdintxt_len)
    i += 1
    ti -= 3

# Done!
