/*
 * Enigma: an interpreter for the Enigma language
 * Copyright (C) 2010  Niels Serup

 * This file is part of Enigma.

 * Enigma is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Enigma is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */

package enigma;

import java.io.*;
import java.util.*;
import java.math.*;
import java.lang.reflect.Method;

/**
 * The class containing the builtin functions used by EnigmaLogic
 * @author  Niels Serup <ns@metanohi.org>
 * @version 0.1
 */
public class EnigmaBuiltinMethods {
    private EnigmaLogic logic;
    
    public EnigmaBuiltinMethods(EnigmaLogic logic) {
        this.logic = logic;
    }
    
    /**
     * Puts variables in a list
     * @param var variable
     * @return    a list with var if var is not a list,
     *            or the list from var if var is a list
     */
    public ArrayList<EnigmaVariable> toList(EnigmaVariable var) {
        if (var.isList())
            return var.getList();
        else {
            ArrayList<EnigmaVariable> n = new ArrayList<EnigmaVariable>();
            n.add(var);
            return n;
        }
    }
    
    /**
     * Decides whether to create a new list value or take an existing
     * non-list value
     * @param arr list of variables
     * @return    a new list value with arr if the size of arr exceeds 1,
     *            or the first item in arr if the size of arr equals 1
     */
    private EnigmaVariable decide(ArrayList<EnigmaVariable> arr) {
        if (arr.size() == 1)
            return arr.get(0);
        else
            return new EnigmaVariable(arr, EnigmaVariableType.LIST);
    }
    
    /**
     * Converts to strings
     * @param arr list of variables
     * @return    string values
     */
    public EnigmaVariable str(ArrayList<EnigmaVariable> arr) {
    for (EnigmaVariable x : arr) {
        x.setValue(x.toString());
        x.setType(EnigmaVariableType.STRING);
    }
    return decide(arr);
    }
    
    /**
     * Converts to numbers
     * @param arr list of variables
     * @return    number values
     */
    public EnigmaVariable num(ArrayList<EnigmaVariable> arr) {
        for (EnigmaVariable x : arr) {
            if (x.isBoolean() || x.isNone()) {
                if (x.isBoolean() && x.getBoolean())
                    x.setValue(new BigDecimal(1));
                else
                    x.setValue(new BigDecimal(0));
                x.setType(EnigmaVariableType.NUMBER);
            }
            else {
                try {
                    x.setValue(new BigDecimal(x.getValue().toString()));
                    x.setType(EnigmaVariableType.NUMBER);
                } catch (Exception e) {
                    logic.putError("num: Value " + x.toRepresentationString()
                            + " cannot be converted to a number.");
                }
            }
        }
        return decide(arr);
    }
    
    /**
     * Puts elements in a list and destroys nested lists at one level
     * @param arr list of variables
     * @return    list values
     */
    public EnigmaVariable list(ArrayList<EnigmaVariable> arr) {
        ArrayList<EnigmaVariable> nArr = new ArrayList<EnigmaVariable>();
        for (EnigmaVariable x : arr) {
            if (x.isList()) {
                for (EnigmaVariable y : x.getList())
                    nArr.add(y);
            }
            else
                nArr.add(x);
        }
        return new EnigmaVariable(nArr, EnigmaVariableType.LIST);
    }
    
    /**
     * Converts to booleans
     * @param arr list of variables
     * @return    boolean values
     */
    public EnigmaVariable bool(ArrayList<EnigmaVariable> arr) {
        boolean arrBool;
        BigDecimal cdec = BigDecimal.ZERO;

        for (EnigmaVariable x : arr) {
            arrBool = (!x.isNone() && (x.isBuiltinCode() || x.isFile() || 
                      ((x.isCode() || x.isString()) && x.getString() != "") ||
                      (x.isList() && x.getList().size() > 0) ||
                      (x.isNumber() && x.getNumber().compareTo(cdec) != 0) ||
                      (x.isBoolean() && x.getBoolean())));
            x.setValue(arrBool);
            x.setType(EnigmaVariableType.BOOLEAN);
        }
        return decide(arr);
    }
    
    /**
     * Converts to code strings
     * @param arr list of variables
     * @return    code values
     */
    public EnigmaVariable code(ArrayList<EnigmaVariable> arr) {
        for (EnigmaVariable x : arr) {
            x.setValue(x.toString());
            x.setType(EnigmaVariableType.CODE);
        }
        return decide(arr);
    }
    
    /**
     * Converts to representational strings
     * @param arr list of variables
     * @return    string values
     */
    public EnigmaVariable repr(ArrayList<EnigmaVariable> arr) {
        for (EnigmaVariable x : arr) {
            x.setValue(x.toRepresentationString());
            x.setType(EnigmaVariableType.STRING);
        }
        return decide(arr);
    }
    
    /**
     * Converts to type strings
     * @param arr list of variables
     * @return    string values
     */
    public EnigmaVariable type(ArrayList<EnigmaVariable> arr) {
        for (EnigmaVariable x : arr) {
            x.setValue(x.getTypeString());
            x.setType(EnigmaVariableType.STRING);
        }
        return decide(arr);
    }
    
    /**
     * Converts to length numbers
     * @param arr list of variables
     * @return    number values
     */
    public EnigmaVariable len(ArrayList<EnigmaVariable> arr) {
        for (EnigmaVariable x : arr) {
            if (x.isBuiltinCode() || x.isNone())
                x.setValue(new BigDecimal(0));
            else if (x.isBoolean()) {
                if (x.getBoolean())
                    x.setValue(new BigDecimal(1));
                else
                    x.setValue(new BigDecimal(0));
            }
            else if (x.isList())
                x.setValue(new BigDecimal(x.getList().size()));
            else
                x.setValue(new BigDecimal(x.toString().length()));
            
            x.setType(EnigmaVariableType.NUMBER);
        }
        return decide(arr);
    }

    /**
     * Clones values
     * @param arr list of variables
     * @return    values
     */
    public EnigmaVariable clone(ArrayList<EnigmaVariable> arr) {
        ArrayList<EnigmaVariable> nArr = new ArrayList<EnigmaVariable>();
        Object value;
        for (EnigmaVariable x : arr) {
            if (!x.isList()) {
                if (x.isString() || x.isCode())
                    value = new String(x.getString());
                else if (x.isNumber())
                    value = new BigDecimal(x.getNumber().toString());
                else if (x.isBoolean())
                    value = new Boolean(x.getBoolean());
                else
                    value = x.getValue();
                nArr.add(new EnigmaVariable(value, x.getType()));
            }
            else {
                nArr.add(clone(x.getList()));
            }
        }
        return decide(nArr);
    }

    /**
     * Slices a string or a list
     * @param arr list of variables
     * @return    a list value sliced according to arr[1] and arr[2]
     *            if arr[0] is a list variable, or substring(
     *            arr[0], arr[1], arr[2]) if arr[0] is a string variable
     */
    public EnigmaVariable slice(ArrayList<EnigmaVariable> arr) {
        int start, end;
        if (arr.size() == 0) {
            logic.putError("slice: There are no variables to slice.");
            return logic.vNone();
        }
        
        EnigmaVariable main = arr.get(0);
        arr.remove(0);
        
        if (arr.size() > 0) {
            try {
                start = arr.get(0).getNumber().intValue();
                arr.remove(0);
            } catch (Exception e) {
                start = 0;
                logic.putError("slice: Start argument is not a number.");
            }
        } else
            start = 0;
        if (arr.size() > 0) {
            try {
                end = arr.get(0).getNumber().intValue();
            } catch (Exception e) {
                end = -1;
                logic.putError("slice: End argument is not a number.");
            }
        } else
            end = -1;
        
        if (main.isString()) {
            String mainString = main.getString();
            if (end == -1)
                end = mainString.length();
            try {
                main.setValue(mainString.substring(start, end));
            } catch (StringIndexOutOfBoundsException e) {
                main.setValue("");
            }
        } else if (main.isList()) {
            ArrayList<EnigmaVariable> mainList = main.getList();
            if (end == -1)
                end = mainList.size();
            end = mainList.size() - end;
            while (start > 0) {
                try {mainList.remove(0);} catch (Exception e) {}
                start--;
                end--;
            }
            while (end > 0) {
                try {mainList.remove(mainList.size() - 1);}
                catch (Exception e) {}
                end--;
            }
            main.setValue(mainList);
        } else
            logic.putError("slice: Value " + main.toRepresentationString()
                    + " cannot be sliced.");
        return main;
    }
    
    /**
     * Runs the same function until it returns false
     * @param arr list of variables
     * @return    none value
     */
    public EnigmaVariable loop(ArrayList<EnigmaVariable> arr) 
    throws Exception {
        if (arr.size() == 0)
            return logic.vNone();
        
        EnigmaVariable code = arr.get(0);
        if (!code.isExecutable()) {
            logic.putError("loop: Value " + code.toRepresentationString()
                    + " is not executable.");
            return logic.vNone();
        }
        
        arr.remove(0);
        EnigmaVariableDictList vars = logic.currentVars;
        ArrayList<EnigmaVariable> arrArgs = new ArrayList<EnigmaVariable>();
        EnigmaVariable returnVariable = null;
        do {
            arrArgs.clear();
            returnVariable = logic.runMethod(code, vars, arr);
            arrArgs.add(returnVariable);
        } while (bool(arrArgs).getBoolean());
        return logic.vNone();
    }
    
    /**
     * Opens a file for reading if arr[1] is "r", writing if arr[1] is "w"
     * or appending if arr[1] is "a"
     * @param arr list of variables
     * @return    the new file value
     */
    public EnigmaVariable open(ArrayList<EnigmaVariable> arr) {
        if (arr.size() == 0) {
            logic.putError("open: There is no filename.");
            return logic.vNone();
        }
        
        EnigmaVariable nameVar = arr.get(0);
        if (!nameVar.isString()) {
            logic.putError("open: Value " + nameVar.toRepresentationString()
                    + " is not a string.");
            return logic.vNone();
        }
        String name = nameVar.getString();
        
        EnigmaVariable flagVar;
        String flag;
        if (arr.size() > 1) {
            flagVar = arr.get(1);
            if (flagVar.isString())
                flag = flagVar.getString();
            else {
                logic.putError("open: Value " +
                        flagVar.toRepresentationString() +
                        " is not a string.");
                flag = "r";
            }
        }
        else
            flag = "r";
        
        EnigmaVariable obj;
        try {
            if (flag.equals("w"))
                obj = new EnigmaVariable(new FileOutputStream(name),
                        EnigmaVariableType.FILE);
            else if (flag.equals("a"))
                obj = new EnigmaVariable(new FileOutputStream(name, true),
                        EnigmaVariableType.FILE);
            else
                obj = new EnigmaVariable(new FileInputStream(name),
                        EnigmaVariableType.FILE);
                
        } catch (IOException e) {
            logic.putError("A problem occured with the file " +
                    nameVar.toRepresentationString() + ".");
            return logic.vNone();
        }
        
        return obj;
    }
    
    /**
     * Closes files
     * @param arr list of variables
     * @return    none value
     */
    public EnigmaVariable close(ArrayList<EnigmaVariable> arr)
    throws IOException {
        for (EnigmaVariable x : arr) {
            if (x.isFile()) {
                try {
                    ((InputStream) x.getValue()).close();
                } catch (Exception e1) {
                    try {
                        ((OutputStream) x.getValue()).close();
                    } catch (Exception e2) {
                        
                    }
                }
            }
            else
                logic.putError("close: Value" + x.toRepresentationString()
                    + " is not a file.");
        }
        return logic.vNone();
    }
        
    /**
     * Reads one character from a file
     * @param arr list of variables
     * @return    string value with read character if character present,
     *            or "" if EOF
     */
    public EnigmaVariable read(ArrayList<EnigmaVariable> arr)
    throws IOException {
        if (arr.size() == 0)
            logic.putError("read: There is no file.");
        else {
            EnigmaVariable first = arr.get(0);
            if (first.isFile()) {
                try {
                    InputStream in = (InputStream) arr.get(0).getValue();
                    arr.remove(0);
                    int[] intArgs = { in.read() };
                    String intString;
                    if (intArgs[0] != -1)
                        intString = new String(intArgs, 0, 1);
                    else
                        intString = "";
                    
                    return new EnigmaVariable(intString, 
                            EnigmaVariableType.STRING);
                } catch (Exception e) {
                    logic.putError("read: File " +
                            arr.get(0).toRepresentationString() + 
                            " is not readable.");
                }
            }
            else
                logic.putError("read: Value " + first.toRepresentationString()
                        + " is not a file.");
        }
        
        return logic.vNone();   
    }
    
    /**
     * Writes string values arr[1...n] separated with spaces to file arr[0]
     * @param arr list of variables
     * @return    none
     */
    public EnigmaVariable write(ArrayList<EnigmaVariable> arr)
    throws IOException {
        if (arr.size() == 0)
            logic.putError("write: There is no file.");
        else {
            EnigmaVariable first = arr.get(0);
            if (first.isFile()) {
                OutputStream out;
                try {
                    out = (PrintStream) arr.get(0).getValue();
                } catch (ClassCastException e1) {
                    try {
                        out = (FileOutputStream) arr.get(0).getValue();
                    } catch (ClassCastException e2) {
                        logic.putError("write: File " +
                                arr.get(0).toRepresentationString() +
                                " does not accept writes.");
                        return logic.vNone();
                    }
                }
                arr.remove(0);
                if (arr.size() == 1 && arr.get(0).isList()) {
                    arr.addAll(arr.get(0).getList());
                    arr.remove(0);
                }
                String text = "";
                for (EnigmaVariable x : arr)
                    text += " " + x;
                if (arr.size() > 0)
                    text = text.substring(1);
                for (int i = 0; i < text.length(); i++)
                    out.write(text.codePointAt(i));
                out.flush();
            }
            else
                logic.putError("write: Value " +
                        first.toRepresentationString() +
                        " is not a file.");
        }
        
        return logic.vNone();
    }
    
    /**
     * Examines if number value arr[0] x arr[1] ... x arr[n] (x = >|<)
     * @param arr         list of variables
     * @param greaterComp greater or lesser comparison
     * @return            true or false, depending on factors
     */
    private EnigmaVariable greaterOrLesser(
            ArrayList<EnigmaVariable> arr, boolean greaterComp) {
        if (arr.size() < 2)
            return logic.vTrue();

        EnigmaVariable x, y;
        boolean good;
        for (int i = 0; i < arr.size() - 1; i++) {
            x = arr.get(i);
            y = arr.get(i + 1);
            if (!x.isNumber()) {
                String gls;
                if (greaterComp)
                    gls = "greater: ";
                else
                    gls = "lesser: ";
                logic.putError(gls + "Value " + x.toRepresentationString()
                        + " is not a number.");
                continue;
            }
            if (greaterComp)
                good = (x.getNumber().compareTo(y.getNumber()) > 0);
            else
                good = (x.getNumber().compareTo(y.getNumber()) < 0);
            if (!good)
                return logic.vFalse();
        }

        // Else
        return logic.vTrue();
    }
    
    /**
     * Examines if number value arr[0] > arr[1] ... > arr[n]
     * @param arr list of variables
     * @return    true if greater than others, false otherwise
     */
    public EnigmaVariable greater(ArrayList<EnigmaVariable> arr) {
        return greaterOrLesser(arr, true);
    }

    /**
     * Examines if number value arr[0] < arr[1] ... < arr[n]
     * @param arr list of variables
     * @return    true if lesser than others, false otherwise
     */
    public EnigmaVariable lesser(ArrayList<EnigmaVariable> arr) {
        return greaterOrLesser(arr, false);
    }
    
    /**
     * Recursively checks if the items in two lists match
     * @param arr1 list of variables
     * @param arr2 list of variables
     * @return    true if matching, false otherwise
     */
    private boolean listsEqual(ArrayList<EnigmaVariable> arr1,
            ArrayList<EnigmaVariable> arr2) {
        if (arr1.size() != arr2.size())
            return false;
        
        EnigmaVariable x, y;
        ArrayList<EnigmaVariable> varList;
        for (int i = 0; i < arr2.size(); i++) {
            x = arr1.get(i);
            y = arr2.get(i);
            if (x.getType() != y.getType())
                return false;
            
            if (!y.isList()) {
                varList = new ArrayList<EnigmaVariable>();
                varList.add(x);
                varList.add(y);
                if (!equal(varList).getBoolean())
                    return false;
            }
            else {
                if (!listsEqual(x.getList(), y.getList()))
                    return false;
            }
        }
        
        return true;
    }
    
    /**
     * Checks if the items of arr match
     * @param arr list of variables
     * @return    true if matching, false otherwise
     */
    public EnigmaVariable equal(ArrayList<EnigmaVariable> arr) {
        if (arr.size() < 2)
            return logic.vTrue();
        
        EnigmaVariable x, y;
        boolean equal;
        for (int i = 0; i < arr.size() - 1; i++) {
            x = arr.get(i);
            y = arr.get(i + 1);
            
            equal = (x.equals(y));
            if (!equal) {
                if (x.isNumber() && y.isNumber())
                    equal = (x.getNumber().compareTo(y.getNumber()) == 0);
                else if (x.isBoolean() && y.isBoolean())
                    equal = (x.getBoolean() == y.getBoolean());
                else if (x.isNone() && y.isNone())
                    equal = true;
                else if ((x.isString() && y.isString()) ||
                        (x.isCode() && y.isCode()))
                    equal = (x.getString().equals(y.getString()));
                else if (x.isList() && y.isList()) {
                    equal = (listsEqual(x.getList(), y.getList()));
                }
            }
            
            if (!equal)
                return logic.vFalse();
        }
        // Else
        return logic.vTrue();
    }
    
    /**
     * Sees if items of arr match the AND paradigm
     * @param arr list of variables
     * @return    true if AND matching, false otherwise
     */
    public EnigmaVariable and(ArrayList<EnigmaVariable> arr) {
        ArrayList<EnigmaVariable> nArr = toList(bool(toList(clone(arr))));
        if (nArr.size() == 1)
            return nArr.get(0);
        else if (nArr.size() == 0)
            return logic.vFalse();

        EnigmaVariable x, y;
        for (int i = 0; i < nArr.size() - 1; i++) {
            x = arr.get(i);
            y = arr.get(i + 1);
            if (!(x.getBoolean() && y.getBoolean()))
                return logic.vFalse();
        }

        // Else
        return logic.vTrue();
    }
    
    /**
     * Sees if items of arr match the OR paradigm
     * @param arr list of variables
     * @return    true if OR matching, false otherwise
     */
    public EnigmaVariable or(ArrayList<EnigmaVariable> arr) {
        ArrayList<EnigmaVariable> nArr = toList(bool(toList(clone(arr))));
        if (nArr.size() == 0)
            return logic.vFalse();

        for (EnigmaVariable x : nArr) {
            if (x.getBoolean())
                return logic.vTrue();
        }

        // Else
        return logic.vFalse();
    }
    
    /**
     * Oppositifies items of arr
     * @param arr list of variables
     * @return    the opposites
     */
    public EnigmaVariable not(ArrayList<EnigmaVariable> arr) {
        arr = toList(bool(arr));
        if (arr.size() == 0)
            return logic.vTrue();
        else {
            for (EnigmaVariable x : arr) {
                x.setValue(new Boolean(!x.getBoolean()));
            }
        }
        return decide(arr);
    }
    
    /**
     * Runs code arr[1] with arguments arr[2...n] if arr[0] is true
     * @param arr list of variables
     * @return    the arr[1] return value
     */
    public EnigmaVariable act(ArrayList<EnigmaVariable> arr)
    throws Exception {
        if (arr.size() < 2)
            return logic.vNone();
        
        EnigmaVariable condition, code;
        condition = arr.get(0);
        arr.remove(0);
        if (!condition.isBoolean()) {
            logic.putError("act: Value " + condition.toRepresentationString()
                    + " is not a boolean.");
            return logic.vNone();
        }
        
        if (!condition.getBoolean())
            return logic.vNone();
        
        code = arr.get(0);
        arr.remove(0);
        if (!code.isExecutable()) {
            logic.putError("act: Value " + code.toRepresentationString()
                    + " is not executable.");
            return logic.vNone();
        }
        
        // Else
        return logic.runMethod(code, logic.currentVars, arr);
    }
    
    /**
     * Issues a system call with arr values
     * @param arr list of variables
     * @return    a list variable with stdout and stderr
     */
    public EnigmaVariable system(ArrayList<EnigmaVariable> arr)
    throws IOException {
        String text = "";
        for (EnigmaVariable x : arr)
            text += " " + x;
        if (arr.size() > 0)
            text = text.substring(1);
        
        Process p = Runtime.getRuntime().exec(text);
        String outText, errText;
        outText = "";
        errText = "";
        
        BufferedReader out = new BufferedReader(
                new InputStreamReader(p.getInputStream()));
        BufferedReader err = new BufferedReader(
                new InputStreamReader(p.getErrorStream()));
        
        while ((text = out.readLine()) != null)
            outText += text + "\n";
        while ((text = err.readLine()) != null)
            errText += text + "\n";
        
        ArrayList<EnigmaVariable> list = new ArrayList<EnigmaVariable>();
        list.add(new EnigmaVariable(outText, EnigmaVariableType.STRING));
        list.add(new EnigmaVariable(errText, EnigmaVariableType.STRING));
        return new EnigmaVariable(list, EnigmaVariableType.LIST);
    }
    
    /**
     * Adds numbers, strings, code or lists together
     * @param arr list of variables
     * @return    the final material
     */
    public EnigmaVariable add(ArrayList<EnigmaVariable> arr) {
        if (arr.size() == 0)
            return logic.vNone();
        
        EnigmaVariable first = arr.get(0);
        arr.remove(0);
        byte type;
        String string = null;
        BigDecimal number = null;
        ArrayList<EnigmaVariable> list = null;
        if (first.isString()) {
            type = 1;
            string = first.getString();
        }
        else if (first.isNumber()) {
            type = 2;
            number = first.getNumber();
        }
        else if (first.isList()) {
            type = 3;
            list = first.getList();
        }
        else if (first.isCode()) {
            type = 4;
            string = first.getString();
        }
        else {
            logic.putError("add: Value " + first.toRepresentationString()
                    + " cannot be added with other objects.");
            return logic.vNone();
        }

        for (EnigmaVariable x : arr) {
            if (type == 1 && (x.isString() || x.isCode()))
                string += x;
            else if (type == 2 && x.isNumber())
                number = number.add(x.getNumber());
            else if (type == 3 && x.isList())
                list.addAll(x.getList());
            else if (type == 4 && (x.isString() || x.isCode()))
                string += x;
            else
                logic.putError("add: Value " + x.toRepresentationString()
                        + " cannot be added with other objects.");
        }
        if (type == 1 || type == 4)
            first.setValue(string);
        else if (type == 2)
            first.setValue(number);
        else if (type == 3)
            first.setValue(list);
        
        return first;
    }
    
    /**
     * Performs the same basic calculation on all elements in arr
     * @param arr  list of variables
     * @param type 1 for subtract, 2 for multiply, 3 for divide
     * @return     the final number
     */
    private EnigmaVariable simpleCalculate(ArrayList<EnigmaVariable> arr,
            byte type) {
        if (arr.size() == 0)
            return logic.vNone();
        
        String typeString;
        if (type == 1)
            typeString = "subtract";
        else if (type == 2)
            typeString = "multiply";
        else
            typeString = "divide";
        
        EnigmaVariable numberVar = null;
        BigDecimal number = null;
        if (arr.get(0).isNumber()) {
            numberVar = arr.get(0);
            number = numberVar.getNumber();
            arr.remove(0);
        } else {
            logic.putError(typeString + ": Value " +
                    arr.get(0).toRepresentationString()
                    + " is not a number.");
            return logic.vNone();
        }
        
        for (EnigmaVariable x : arr) {
            if (x.isNumber()) {
                if (type == 1)
                    number = number.subtract(x.getNumber());
                else if (type == 2)
                    number = number.multiply(x.getNumber());
                else
                    number = number.divide(x.getNumber());
            } else {
                logic.putError(typeString + ": Value " +
                        arr.get(0).toRepresentationString()
                        + " is not a number.");
            }
        }
        numberVar.setValue(number);
        return numberVar;
    }
    
    /**
     * Subtracts all elements from each other
     * @param arr list of variables
     * @return    the final number
     */
    public EnigmaVariable subtract(ArrayList<EnigmaVariable> arr) {
        byte a = 1;
        return simpleCalculate(arr, a);
    }
    
    /**
     * Multiplies all elements with each other
     * @param arr list of variables
     * @return    the final number
     */
    public EnigmaVariable multiply(ArrayList<EnigmaVariable> arr) {
        byte a = 2;
        return simpleCalculate(arr, a);
    }
    
    /**
     * Divides all elements with each other
     * @param arr list of variables
     * @return    the final number
     */
    public EnigmaVariable divide(ArrayList<EnigmaVariable> arr) {
        byte a = 3;
        return simpleCalculate(arr, a);
    }
    
    /**
     * Checks if the first two items of arr are numbers
     * @param arr  list of variables
     * @param name the name of the calling function
     * @return     the number variables if the first two items of arr
     *             are numbers, a none variable otherwise
     */
    private EnigmaVariable[] checkIfTwoNumbers(ArrayList<EnigmaVariable> arr,
            String name) {
        EnigmaVariable[] noneArgs = { logic.vNone() };
        if (arr.size() == 0) {
            logic.putError(name + ": There is no number.");
            return noneArgs;
        }
        EnigmaVariable aVar, bVar;
        aVar = arr.get(0);
        if (!aVar.isNumber()) {
            logic.putError(name + ": Value " + aVar.toRepresentationString()
                    + " is not a number.");
            return noneArgs;
        }
        if (arr.size() > 1) {
            bVar = arr.get(1);
            if (!bVar.isNumber()) {
                logic.putError(name + ": Value " +
                        aVar.toRepresentationString() + " is not a number.");
                return noneArgs;
            }
        }
        else {
            logic.putError(name + ": There is no number.");
            return noneArgs;
        }
        // Else
        EnigmaVariable[] args = { aVar, bVar };
        return args;
    }
    
    /**
     * Finds the remainder of arr[0] / arr[1]
     * @param arr list of variables
     * @return    the remainder
     */
    public EnigmaVariable mod(ArrayList<EnigmaVariable> arr) {
        EnigmaVariable[] vars = checkIfTwoNumbers(arr, "mod");
        if (vars[0].isNone())
            return logic.vNone();
        
        return new EnigmaVariable(vars[0].getNumber().remainder(
                vars[1].getNumber()), EnigmaVariableType.NUMBER);
    }
    
    /**
     * Calculates arr[0]^arr[1]
     * @param arr list of variables
     * @return    the number
     */
    public EnigmaVariable pow(ArrayList<EnigmaVariable> arr) {
        EnigmaVariable[] vars = checkIfTwoNumbers(arr, "pow");
        if (vars[0].isNone())
            return logic.vNone();
        
        BigDecimal number = null;
        try {
            number = vars[0].getNumber().pow(vars[1].getNumber().
                    intValueExact());
        } catch (ArithmeticException e) {
            number = new BigDecimal(Math.pow(vars[0].getNumber().
                    doubleValue(), vars[1].getNumber().doubleValue()));
        }
        return new EnigmaVariable(number, EnigmaVariableType.NUMBER);
    }
    
    /**
     * Calculates log(arr[0], arr[1]) or log(arr[0], e)
     * @param arr list of variables
     * @return    the number
     */
    public EnigmaVariable log(ArrayList<EnigmaVariable> arr) {
        if (arr.size() == 1)
            arr.add(new EnigmaVariable(EnigmaLogic.E,
                    EnigmaVariableType.NUMBER));
        EnigmaVariable[] vars = checkIfTwoNumbers(arr, "log");
        if (vars[0].isNone())
            return logic.vNone();
        
        BigDecimal number = new BigDecimal(
                Math.log(vars[0].getNumber().doubleValue()) /
                Math.log(vars[1].getNumber().doubleValue()));
        return new EnigmaVariable(number, EnigmaVariableType.NUMBER);
    }
    
    /**
     * Generates a random number
     * @param arr list of variables
     * @return    the number
     */
    public EnigmaVariable random(ArrayList<EnigmaVariable> arr) {
        return new EnigmaVariable(new BigDecimal(Math.random()),
                EnigmaVariableType.NUMBER);
    }
    
    /**
     * Finds the absolute value of number variables
     * @param arr list of variables
     * @return    the absolute values
     */
    public EnigmaVariable abs(ArrayList<EnigmaVariable> arr) {
        for (EnigmaVariable x : arr) {
            try {
                x.setValue(x.getNumber().abs());
            } catch (Exception e) {}
        }
        return decide(arr);
    }
    
    /**
     * Rounds arr[0] with arr[1] decimals
     * @param arr          list of variables
     * @param roundingMode what rounding mode to be used
     * @return             the number
     * @see java.math.RoundingMode
     */
    private EnigmaVariable round(ArrayList<EnigmaVariable> arr,
            RoundingMode roundingMode) {
        if (arr.size() == 1)
            arr.add(new EnigmaVariable(BigDecimal.ZERO,
                    EnigmaVariableType.NUMBER));
        EnigmaVariable[] vars = checkIfTwoNumbers(arr, "round");
        if (vars[0].isNone())
            return logic.vNone();
        BigDecimal number = vars[0].getNumber();
        MathContext con = new MathContext(vars[1].getNumber().intValue() +
                number.precision() - number.scale(), roundingMode);
        return new EnigmaVariable(number.round(con),
                EnigmaVariableType.NUMBER);
    }

    /**
     * Rounds arr[0] with arr[1] decimals upwards if .5, otherwise downwards
     * @param arr list of variables
     * @return    the number
     */
    public EnigmaVariable round(ArrayList<EnigmaVariable> arr) {
        return round(arr, RoundingMode.HALF_UP);
    }
    
    /**
     * Rounds arr[0] with arr[1] decimals downwards
     * @param arr list of variables
     * @return    the number
     */
    public EnigmaVariable floor(ArrayList<EnigmaVariable> arr) {
        return round(arr, RoundingMode.FLOOR);
    }
    
    /**
     * Rounds arr[0] with arr[1] decimals upwards
     * @param arr list of variables
     * @return    the number
     */
    public EnigmaVariable ceil(ArrayList<EnigmaVariable> arr) {
        return round(arr, RoundingMode.CEILING);
    }
    
    /**
     * Finds trigonometric values of numbers in arr
     * @param arr        list of variables
     * @param methodName the name of the calling method
     * @return           the new numbers
     */
    private EnigmaVariable trigonometric(ArrayList<EnigmaVariable> arr,
            String methodName) throws Exception {
        Class[] stdArgs = new Class[] { double.class };
        Method method = Math.class.getMethod(methodName, stdArgs);
        for (EnigmaVariable x : arr) {
            Object[] args = { x.getNumber().doubleValue() };
            x.setValue(new BigDecimal((Double) method.invoke(null, args)));
        }
        return decide(arr);
    }
    
    /**
     * Finds sine of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable sin(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "sin");
    }
    
    /**
     * Finds cosine of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable cos(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "cos");
    }
    
    /**
     * Finds tangent of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable tan(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "tan");
    }
    
    /**
     * Finds arcsine of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable asin(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "asin");
    }
    
    /**
     * Finds arccosine of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable acos(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "acos");
    }
    
    /**
     * Finds arctangent of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable atan(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "atan");
    }
    
    /**
     * Finds hyperbolic sine of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable sinh(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "sinh");
    }
    
    /**
     * Finds hyperbolic cosine of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable cosh(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "cosh");
    }
    
    /**
     * Finds hyperbolic tangent of numbers in arr
     * @param arr list of variables
     * @return    the new numbers
     */
    public EnigmaVariable tanh(ArrayList<EnigmaVariable> arr)
    throws Exception {
        return trigonometric(arr, "tanh");
    }
}
