/*
 * Enigma: an interpreter for the Enigma language
 * Copyright (C) 2010  Niels Serup

 * This file is part of Enigma.

 * Enigma is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Enigma is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */

package enigma;

import java.io.*;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.math.BigDecimal;

/**
 * The class containing the logic needed to run Enigma programs
 * @author  Niels Serup <ns@metanohi.org>
 * @version 0.1
 */
public class EnigmaLogic {
    public EnigmaBuiltinMethods builtinMethods;
    public EnigmaVariableDict globals;
    EnigmaVariableDictList globalsArray;    
    EnigmaVariableDictList currentVars;

    public static EnigmaVariable noneVariable = new EnigmaVariable(
            '\0', EnigmaVariableType.NONE);
    public static EnigmaVariable trueVariable = new EnigmaVariable(
            new Boolean(true), EnigmaVariableType.BOOLEAN);
    public static EnigmaVariable falseVariable = new EnigmaVariable(
            new Boolean(false), EnigmaVariableType.BOOLEAN);
    public static EnigmaVariable PI = new EnigmaVariable(
            new BigDecimal("3.1415926535"), EnigmaVariableType.NUMBER);
    public static EnigmaVariable E = new EnigmaVariable(
            new BigDecimal("2.7182818284"), EnigmaVariableType.NUMBER);
    
    public class NullOutputStream extends OutputStream {
        @Override
        public void write(int c) {
        }
    }
    
    boolean showErrors;

    public EnigmaLogic(boolean showErrors) throws Exception {
        this.showErrors = showErrors;

        globals = new EnigmaVariableDict();
        globalsArray = new EnigmaVariableDictList();
        globalsArray.add(globals);

        globals.put("stdout", new EnigmaVariable(System.out,
                EnigmaVariableType.FILE));
        globals.put("stdin", new EnigmaVariable(System.in,
                EnigmaVariableType.FILE));
        globals.put("stderr", new EnigmaVariable(System.err,
                EnigmaVariableType.FILE));
        globals.put("zero", new EnigmaVariable(new PrintStream(
                new NullOutputStream()), EnigmaVariableType.FILE));
        globals.put("none", vNone());
        globals.put("true", vTrue());
        globals.put("false", vFalse());
        globals.put("@pi", new EnigmaVariable(PI));
        globals.put("@e", new EnigmaVariable(E));
        globals.put("cwd", new EnigmaVariable(new File(".").
                getCanonicalPath() + "/", EnigmaVariableType.STRING));

        createReferencesToBuiltinFunctions(
                "str", "num", "list", "bool", "code", "repr", "type", "len",
                "clone", "slice", "loop", "open", "close", "read", "write",
                "greater", "lesser", "equal", "not", "and", "or", "act",
                "system", "add", "subtract", "multiply", "divide", "mod",
                "pow", "log", "random", "abs", "round", "floor", "ceil",
                "sin", "cos", "tan", "asin", "acos", "atan", "sinh",
                "cosh", "tanh"
        );
    }
    
    public EnigmaVariable vNone() {
        return new EnigmaVariable(noneVariable);
    }
    
    public EnigmaVariable vTrue() {
        return new EnigmaVariable(trueVariable);
    }
    
    public EnigmaVariable vFalse() {
        return new EnigmaVariable(falseVariable);
    }

    private String readFile(String filename) throws IOException {
        FileReader file = null;
        try {
            file = new FileReader(filename);
        } catch (FileNotFoundException e) {
            if (showErrors)
                System.err.println("File " + filename + " does not exist.");
            System.exit(1);
        }
        BufferedReader in = new BufferedReader(file);

        String content = "";

        String line = in.readLine();
        while (line != null) {
            content += line + "\n";
            line = in.readLine();
        }
        in.close();
        
        content = content.replaceAll("\r?\n", "\n");
        content = removeComments(content);
        return content;
    }

    public void parseFile(String filename, String[] args)
    throws Exception {
        String allcode = readFile(filename);
        
        String currentProgramDir = new File(filename).getCanonicalPath();
        currentProgramDir = currentProgramDir.substring(0,
                currentProgramDir.lastIndexOf('/') + 1);
        globals.put("cpd", new EnigmaVariable(currentProgramDir,
                EnigmaVariableType.STRING));
        globals.put("fnm", new EnigmaVariable(filename,
                EnigmaVariableType.STRING));

        ArrayList<EnigmaVariable> eArgs = new ArrayList<EnigmaVariable>();
        for (String x : args) {
            eArgs.add(new EnigmaVariable(x, EnigmaVariableType.STRING));
        }
        parseCode(allcode, globalsArray, eArgs);
    }
    

    private void createReferencesToBuiltinFunctions(
            String... methodsNames) throws Exception {
        builtinMethods = new EnigmaBuiltinMethods(this);

        Class[] stdArgs = new Class[] { ArrayList.class };
        for (String x : methodsNames) {
            Method m = EnigmaBuiltinMethods.class.getMethod(x, stdArgs);
            globals.put(x,
                    new EnigmaVariable(m, EnigmaVariableType.BUILTINCODE));
        }
    }
    
    private String removeComments(String code) {
        String newCode = "";
        String[] splitCode = code.split("\n");
        
        int i, c;
        for (String x : splitCode) {
            for (i = 0; i < x.length(); i++) {
                c = x.codePointAt(i);
                if (c == '#') {
                    newCode += "\n";
                    break;
                }
                else if (c != ' ' && c != '\t') {
                    newCode += x + "\n";
                    break;
                }
            }
        }
        return newCode;
    }

    public EnigmaVariable runMethod(EnigmaVariable var,
            EnigmaVariableDictList varsArr, ArrayList<EnigmaVariable> arrArgs)
            throws Exception {
        if (var == null || !var.isExecutable()) {
            putError("Value " + var.toRepresentationString()
                    + " is not executable.");
            return vNone();
        }
        if (var.isBuiltinCode()) {
            Method me = var.getBuiltinCode();
            Object[] args = { arrArgs };
            EnigmaVariable res = (EnigmaVariable) me.invoke(builtinMethods, args);
            return res;
        } else if (var.isCode())
            return parseCode(var.getString(), varsArr, arrArgs);
        else
            return vNone();
    }
    
    public void putError(String msg) {
        if (showErrors) {
            try {
                ((PrintStream) globals.getVariable("stderr").getValue()).
                println("Error: " + msg);
            } catch (ClassCastException e) {}
        }
    }
    
    public BigDecimal stringToNumber(String test) {
        try {
            return new BigDecimal(test);    
        } catch (NumberFormatException e) {
            return null;
        }
    }
    
    /**
     * This method is the core function in the Enigma interpreter. It reads the
     * code character by character and performs actions based on what it sees.
     * It uses several variables to keep track of what is happening and takes
     * care not to mix different types of variables, as it could become quite a
     * mess.
     * This method creates a new EnigmaVariableDict "below" the current ones.
     * It exists only until this method ends.
     * @param origCode code string
     * @param origVars current list of variable dicts in hierarchies
     * @param arrArgs  arguments to the code
     * @return         return value if specified in code, none value otherwise
     */
    public EnigmaVariable parseCode(String origCode,
            EnigmaVariableDictList origVars,
            ArrayList<EnigmaVariable> arrArgs)
            throws Exception {
        String code = new String(origCode);
        if (code.length() == 0)
            return vNone();
        
        EnigmaVariableDict localDict;
        EnigmaVariable localTempVar;
        boolean useGlobalSpace;
        
        // Copies origVars into vars
        EnigmaVariableDictList vars = new EnigmaVariableDictList();
        for (EnigmaVariableDict x : origVars) {vars.add(x);}
        currentVars = vars; // Sets the public variable to vars for easy access
        
        // Creates a new EnigmaVariableDict to keep the variables of this code
        localDict = vars.addDict();
        localDict.put("args", new EnigmaVariable(arrArgs,
                EnigmaVariableType.LIST));
        localTempVar = vNone(); // Initiates "temp"
        localDict.put("temp", localTempVar);

        int position;
        
        if (code.codePointAt(0) == '*') {
            useGlobalSpace = true;
            code = code.substring(1);
        }
        else
            useGlobalSpace = false;

        // Sees if there are any argument names enclosed in /.../ present
        position = 0;
        if (code.codePointAt(0) == '/') {
            String[] argsNames;

            position = code.indexOf('/', 1);
            argsNames = code.substring(1, position).split("(,|\\s+)");
            code = code.substring(position + 1);
            position = 0;
            int size = arrArgs.size();
            for (String x : argsNames) {
                if (position < size) {
                    localDict.put(x.trim(), arrArgs.get(position));
                } else {
                    localDict.put(x.trim(), vNone());
                }
                position++;
            }
        }
        position = 0;
        
        // Defines several variables to be used in the upcoming loop
        String tempContent;
        ArrayList<ArrayList<EnigmaVariable>> tempVars;
        int tempArrNum;
        ArrayList<EnigmaVariable> tempArrEntry;

        byte currentVariableType;
        byte currentActionType;
        String currentCharString, previousCharString, nextCharString;
        int currentChar, previousChar, nextChar;
        boolean setVarSign, setGlobalVarSign,
        actionSign, endSign, semiEndSign;

        EnigmaVariable tempUsableVar;
        ArrayList<Byte> insideCodeType;
        byte insideCodeTypeLatest;

        // Initiates the variables
        tempContent = "";
        tempVars = new ArrayList<ArrayList<EnigmaVariable>>();
        tempVars.add(new ArrayList<EnigmaVariable>());
        tempArrNum = 0;
        tempArrEntry = tempVars.get(tempArrNum);

        currentVariableType = 0;
        currentActionType = 0;

        insideCodeTypeLatest = 0;
        insideCodeType = new ArrayList<Byte>();
        insideCodeType.add(Byte.valueOf(insideCodeTypeLatest));

        // Runs the actual loop. This is where the fun happens.
        while (position < code.length()) {
            // First, find the current, the previous and the next characters
            currentCharString = code.substring(position, position + 1);
            currentChar = currentCharString.codePointAt(0);
            try {
                previousCharString = code.substring(position - 1, position);
                previousChar = previousCharString.codePointAt(0);
            } catch (StringIndexOutOfBoundsException e) {
                previousCharString = "";
                previousChar = 0;
            }
            try {
                nextCharString = code.substring(position + 1, position + 2);
                nextChar = nextCharString.codePointAt(0);
            } catch (StringIndexOutOfBoundsException e) {
                nextCharString = "";
                nextChar = 0;
            }
            
            // Second, look at the value of currentVariableType
            if (currentVariableType == 0) {
                /* This is the type that all statements begin with. After a
                 * value has "ended", this also becomes the new type.
                 */
                setVarSign = (currentChar == '=');
                setGlobalVarSign = (currentChar == '*');
                actionSign = (currentChar == '!');
                semiEndSign = (currentChar == '|');
                endSign = (currentChar == ';');

                if (currentChar == ' ' || currentChar == '\n'
                        || currentChar == '\t' || currentChar == ',') {
                    // Nothing
                    position++;
                } else if (setVarSign || setGlobalVarSign || actionSign ||
                        semiEndSign || endSign) {
                    // Actions
                    if (currentActionType == '=' || currentActionType == '*') {
                        // Set the value of old variables to new variables
                        if (tempVars.get(tempArrNum - 1).size() == 1)
                            tempUsableVar = tempVars.get(
                                    tempArrNum - 1).get(0);
                        else
                            tempUsableVar = new EnigmaVariable(
                                    tempVars.get(tempArrNum - 1),
                                    EnigmaVariableType.LIST);
                        if (tempArrEntry.size() == 0)
                            putError("There is no variable to " +
                                    "save the variables in.");
                        for (EnigmaVariable x : tempArrEntry) {
                            if (x.isNone()) {
                                if (useGlobalSpace || currentActionType == '*')
                                    globals.put(x.getString(), tempUsableVar);
                                else
                                    localDict.put(x.getString(), tempUsableVar);
                            }
                            else
                                x.setVariable(tempUsableVar);
                        }
                        tempVars.remove(tempArrNum);
                        tempVars.add(tempArrNum, 
                                tempVars.get(tempArrNum - 1));
                    } else if (currentActionType == '!') {
                        // Send data to code (methods)
                        ArrayList<EnigmaVariable> tempMethodArgs;
                        tempMethodArgs = tempVars.get(tempArrNum - 1);
                        if (tempMethodArgs.size() == 1
                                && tempMethodArgs.get(0).isList())
                            tempMethodArgs = tempMethodArgs.get(0).getList();
                        if (tempArrEntry.size() == 0)
                            putError("There is no code to execute.");
                        else {
                            // The data can be sent to several functions
                            for (EnigmaVariable x : tempArrEntry) {
                                tempUsableVar = runMethod(x, vars,
                                        tempMethodArgs);
                                tempMethodArgs = new ArrayList<EnigmaVariable>();
                                tempMethodArgs.add(tempUsableVar);
                            }
                            tempArrEntry.clear();
                            if (tempMethodArgs.size() == 1 && tempMethodArgs.get(0).isList())
                                tempMethodArgs = tempMethodArgs.get(0).getList(); 
                            tempArrEntry.addAll(tempMethodArgs);
                        }
                    }

                    if (setVarSign || setGlobalVarSign || actionSign) {
                        // If a part has ended
                        tempArrNum++;
                        tempVars.add(new ArrayList<EnigmaVariable>());
                        tempArrEntry = tempVars.get(tempArrNum);
                    }

                    // Define new actions
                    if (setVarSign)
                        currentActionType = '=';
                    else if (setGlobalVarSign)
                        currentActionType = '*';
                    else if (actionSign)
                        currentActionType = '!';
                    else if (endSign || (semiEndSign &&
                            currentActionType == '!')) {
                        // In case of a b c ! x | ...
                        if (semiEndSign && currentActionType == '!') {
                            // Modify the "temp" variable
                            if (tempArrEntry.size() == 1)
                                tempUsableVar = tempArrEntry.get(0);
                            else
                                tempUsableVar = new EnigmaVariable(
                                        tempArrEntry, EnigmaVariableType.LIST);
                            localTempVar.setVariable(tempUsableVar);
                        }
                        // Start over with a new statement
                        tempContent = "";
                        tempVars.clear();
                        tempVars.add(new ArrayList<EnigmaVariable>());
                        tempArrNum = 0;
                        tempArrEntry = tempVars.get(tempArrNum);
                        
                        currentVariableType = 0;
                        currentActionType = 0;
                    }
                    position++;
                } else {
                    // Ready for input
                    tempContent = "";
                    if (currentChar == '{') {
                        currentVariableType = 1;
                        position++;
                    } else if (currentChar == '"') {
                        currentVariableType = 2;
                        position++;
                    } else
                        currentVariableType = 3;
                }
            } else if (currentVariableType == 1) {
                // A code object
                /* Since code objects can contain both strings and other code
                 * strings, it is a must to have a list containing what is
                 * currently the case, as ending a code object cannot be done
                 * correctly without. Simply looking for a '}' might otherwise
                 * lead to one inside a string inside the code object.
                 */
                if (insideCodeTypeLatest == 0) {
                    if (currentChar == '}') {
                        // End it
                            currentVariableType = 0;
                        if (currentActionType == '=' ||
                                currentActionType == '*')
                            putError("A piece of code (" +
                                    new EnigmaVariable(tempContent,
                                    EnigmaVariableType.CODE).
                                    toRepresentationString() +
                                                                        ") cannot be used as a variable name.");
                                                else
                                                        tempArrEntry.add(new EnigmaVariable(tempContent,
                                                                        EnigmaVariableType.CODE));
                                                position++;
                                                continue;
                                        } else if (currentChar == '{') {
                                                // A code object in the code object begins
                                                insideCodeTypeLatest = '{';
                                                insideCodeType.add(Byte.valueOf(
                                                                insideCodeTypeLatest));
                                        } else if (currentChar == '"') {
                                                // A string object in the code object begins
                                                insideCodeTypeLatest = '"';
                                                insideCodeType.add(Byte.valueOf(
                                                                insideCodeTypeLatest));
                                        }
                                } else if (insideCodeTypeLatest == '{') {
                                        // Current type is a code object
                                        if (currentChar == '"') {
                                                // A string object in the code object begins
                                                insideCodeTypeLatest = '"';
                                                insideCodeType.add(Byte.valueOf(
                                                                insideCodeTypeLatest));
                                        } else if (currentChar == '}') {
                                                // The code object ends
                                                int num = insideCodeType.size() - 1;
                                                insideCodeType.remove(num);
                                                insideCodeTypeLatest = insideCodeType.get(num - 1).
                                                byteValue();
                                        } else if (currentChar == '{') {
                                            // A new code object begins
                                            insideCodeType.add(Byte.valueOf(
                                                        insideCodeTypeLatest));
                                        }
                                } else if (insideCodeTypeLatest == '"') {
                                        // Current type is a string object
                                        if (currentChar == '"' && previousChar != '\\') {
                                                // The string object ends               
                                                int num = insideCodeType.size() - 1;
                                                insideCodeType.remove(num);
                                                insideCodeTypeLatest = insideCodeType.get(num - 1).
                                                byteValue();
                                        }
                                }
                                tempContent += currentCharString;
                                position++;
                        } else if (currentVariableType == 2) {
                                // A string object
                                if (currentChar == '\\') {
                                        // Common escape sequences
                                        if (nextChar == 'n')
                                                tempContent += "\n";
                                        else if (nextChar == 't')
                                                tempContent += "\t";
                                        else if (nextChar != '\n')
                                                tempContent += nextCharString;
                                        position += 2;
                                } else if (currentChar == '"') {
                                        // End the string object
                                        currentVariableType = 0;
                                        if (currentActionType == '!')
                                                putError("A string (" + new EnigmaVariable(
                                                                tempContent, EnigmaVariableType.STRING).
                                                                toRepresentationString() +
                                                                ") cannot be used as a code object.");
                                        else if (currentActionType == '=' ||
                                                        currentActionType == '*')
                                                putError("A string (" + new EnigmaVariable(
                                                                tempContent, EnigmaVariableType.STRING).
                                                                toRepresentationString() +
                                                                ") cannot be used as a variable name.");
                                        else
                                                tempArrEntry.add(new EnigmaVariable(tempContent,
                                                                EnigmaVariableType.STRING));
                                        position++;
                                } else {
                                        tempContent += currentCharString;
                                        position++;
                                }
                        } else if (currentVariableType == 3) {
                                // A number or an id for an already existing variable
                                if (currentChar == ' ' || currentChar == '\t'
                                                || currentChar == '\n' || currentChar == ','
                                                || currentChar == '=' || currentChar == '*' 
                                                || currentChar == '!' || currentChar == '|'
                                                || currentChar == ';') {
                                        // End it
                                        currentVariableType = 0;

                                        // See if it's a number or an old variable or none
                                        BigDecimal stringNumber = stringToNumber(tempContent);
                                        if (currentActionType == 0) {
                                                if (stringNumber != null) {
                                                        tempArrEntry.add(new EnigmaVariable(stringNumber,
                                                                        EnigmaVariableType.NUMBER));
                                                } else {
                                                        tempUsableVar = vars.getVariable(tempContent);
                                                        if (tempUsableVar != null)
                                                                tempArrEntry.add(tempUsableVar);
                                                        else
                                                                putError("Variable \"" + tempContent
                                                                                + "\" does not exist.");
                                                }
                                        } else if (currentActionType == '!') {
                                                tempUsableVar = vars.getVariable(tempContent);
                                                if (tempUsableVar != null) {
                                                        if (tempUsableVar.isExecutable())
                                                                tempArrEntry.add(tempUsableVar);
                                                        else
                                                                putError("Variable \"" + tempContent
                                                                                + "\" is not a code object.");
                                                }
                                                else
                                                        putError("Variable \"" + tempContent
                                                                        + "\" does not exist.");
                                        } else if (currentActionType == '=' ||
                                                        currentActionType == '*') {
                                                tempUsableVar = vars.getVariable(tempContent);
                                                if (tempUsableVar != null) {
                                                        tempArrEntry.add(tempUsableVar);
                                                } else if (stringNumber == null) {
                                                        tempUsableVar = new EnigmaVariable(
                                                                        tempContent, EnigmaVariableType.NONE);
                                                        tempArrEntry.add(tempUsableVar);
                                                } else 
                                                        putError("A number (" + new EnigmaVariable(
                                                                stringNumber, EnigmaVariableType.NUMBER).
                                                                toRepresentationString()
                                                                + ") cannot be used as a variable name.");
                                        }
                                } else {
                                        // Add the current character to the temporary string
                                        tempContent += currentCharString;
                                        position++;
                }
            }
        }
        
        // Attempt to get the return value
        EnigmaVariable returnValue = localDict.getVariable("return");
        if (returnValue == null)
            returnValue = vNone();
        return returnValue;
    }
}
