/*
 * Enigma: an interpreter for the Enigma language
 * Copyright (C) 2010  Niels Serup

 * This file is part of Enigma.

 * Enigma is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Enigma is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */

package enigma;

import java.util.ArrayList;
import java.math.BigDecimal;
import java.lang.reflect.Method;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * The variable class used throughout the Enigma interpreter
 * @author  Niels Serup <ns@metanohi.org>
 * @version 0.1
 */
public class EnigmaVariable {
    private Object content;

    private EnigmaVariableType contentType;

    public EnigmaVariable(Object value, EnigmaVariableType type) {
        setValue(value);
        setType(type);
    }
    public EnigmaVariable(EnigmaVariable var) {
        setVariable(var);
    }
    
    public void setValue(Object value) {
        content = value;
    }
    
    public void setType(EnigmaVariableType type) {
        contentType = type;
    }
    
    public void setVariable(EnigmaVariable var) {
        setValue(var.getValue());
        setType(var.getType());
    }
    
    public Object getValue() {
        return content;
    }
    
    public EnigmaVariableType getType() {
        return contentType;
    }
    
    public boolean hasType(EnigmaVariableType type) {
        return type == getType();
    }
    
    public boolean isCode() {
        return (hasType(EnigmaVariableType.CODE));
    }
    public String getCode() {
        return getString();
    }
    
    public boolean isBuiltinCode() {
        return (hasType(EnigmaVariableType.BUILTINCODE));
    }
    public Method getBuiltinCode() {
        return (Method) getValue();
    }
    
    public boolean isExecutable() {
        return (isCode() || isBuiltinCode());
    }
    
    public boolean isList() {
        return (hasType(EnigmaVariableType.LIST));
    }
    @SuppressWarnings("unchecked")
    public ArrayList<EnigmaVariable> getList() {
        return (ArrayList<EnigmaVariable>) getValue();
    }
    
    public boolean isString() {
        return (hasType(EnigmaVariableType.STRING));
    }
    public String getString() {
        return (String) getValue();
    }
    
    public boolean isNumber() {
        return (hasType(EnigmaVariableType.NUMBER));
    }
    public BigDecimal getNumber() {
        return (BigDecimal) getValue();
    }
    
    public boolean isFile() {
        return (hasType(EnigmaVariableType.FILE));
    }
    
    public boolean isBoolean() {
        return (hasType(EnigmaVariableType.BOOLEAN));
    }
    public boolean getBoolean() {
        return ((Boolean) getValue()).booleanValue();
    }
    
    public boolean isNone() {
        return (hasType(EnigmaVariableType.NONE));
    }
    
    public String toString() {
        if (isList()) {
            String text = "";
            for (EnigmaVariable x : getList())
                text += ", " + x.toRepresentationString();
            if (getList().size() > 0)
                text = text.substring(2);
            return "[" + text + "]";
        } else if (isFile()) {
            try {
                @SuppressWarnings("unused")
                InputStream i = (InputStream) getValue();
                return "{file[in]}";
            } catch (Exception e1) {
                try {
                    @SuppressWarnings("unused")
                    OutputStream o = (OutputStream) getValue();
                    return "{file[out]}";
                } catch (Exception e2) {
                    return "{file}";
                }
            }
        } else if (isBuiltinCode()) {
            String complete = content.toString();
            complete = complete.substring(0, complete.indexOf('('));
            complete = complete.substring(complete.lastIndexOf('.') + 1);
            return "{builtincode:" + complete + "}";
        } else if (isBoolean()) {
            return (getBoolean() ? "true" : "false");
        } else if (isNumber()) {
            return ((BigDecimal) content).toPlainString();
        } else if (isNone()) {
            return "none";
        } else
            return content.toString();
    }
    
    public String toRepresentationString() {
        if (isCode())
            return "{" + toString() + "}";
        else if (isString())
            return "\"" + getString().replaceAll("\"", "\\\"").
            replaceAll("\n", "\\\\n").replaceAll("\t", "\\\\t") + "\"";
        else
            return toString();
    }
    
    public String getTypeString() {
        return contentType.toString();
    }
}
