#!/usr/bin/env python
# -*- coding: utf-8 -*-

# vit: a collection of boolean logic tools
# Copyright (C) 2010  Niels Serup

# This file is part of vit.
#
# vit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vit.  If not, see <http://www.gnu.org/licenses/>.

##[ Name        ]## vit.utility
##[ Maintainer  ]## Niels Serup <ns@metanohi.org>
##[ Description ]## Controls the actions of the command-line utility

import os
import datetime
import tempfile
import locale
import atexit
from vit.settingsparser import SettingsParser
from vit.database import Database
import vit.various as various
import vit.generalinformation as ginfo

_config_file_translations = {
    'verbose': 'term_verbose',
    'colored text': 'term_colored_text',
    'editor': 'editor'
}


class Utility(SettingsParser):
    def __init__(self, **options):
        SettingsParser.__init__(self, _config_file_translations,
                                **options)
        # Get values, or set them to default ones
        self.set_if_nil('error_function', various._usable_error)
        self.set_if_nil('status_function', various._usable_status)
        self.set_if_nil('term_verbose', True)
        self.set_if_nil('term_colored_text', True)
        self.set_if_nil('vit_path', '.')
        self.set_if_nil('config_file_path', 'vit.conf')
        self.set_if_nil('database_path', 'vit.sqlite')
        self.set_if_nil('editor', None)

        self.editors = (
            self.editor,
            os.environ['EDITOR'],
            os.environ['VISUAL'],
            'editor',
            'emacs',
            'nano',
            'joe',
            'ed')

        self.vit_path = os.path.abspath(self.vit_path)
        self.config_file_path = os.path.abspath(self.config_file_path)
        self.database_path = os.path.abspath(self.database_path)

    def error(self, msg, done=None):
        if self.term_verbose:
            self.error_function(msg, done, use_color=self.term_colored_text)

    def status(self, msg):
        if self.term_verbose:
            self.status_function(msg, use_color=self.term_colored_text)

    def run_editor(self, fname):
        done = False
        for editor in self.editors:
            if not editor:
                continue
            try:
                various.exec_program(editor, fname)
                done = True
                break
            except OSError:
                pass
        return done

    def create_database(self):
        try:
            self.database.create()
            self.status('database has been created')
        except Exception, e:
            self.error(e)

    def add_entry(self):
        begin_time = datetime.datetime.now()
        tf = tempfile.NamedTemporaryFile()
        if not self.run_editor(tf.name):
            self.error('missing an editor', True)
        end_time = datetime.datetime.now()
        text = tf.file.read()
        if not text:
            self.status('not recording message (empty)')
            return
        try:
            text = text.decode(locale.getdefaultlocale()[1])
        except Exception:
            pass
        try:
            self.database.add(text, begin_time, end_time)
            self.status('message recorded')
        except Exception, e:
            self.error(e)

    def show_list(self):
        try:
            for msg, begin, end in self.database.get():
                print ('<< ' + begin.strftime('%Y-%m-%d %H:%M:%S') +
                       ' -- ' + end.strftime('%Y-%m-%d %H:%M:%S') +
                       ' >>').center(70)
                print '¨' * 70
                print msg
                print '_' * 70
                print '\n'
        except Exception, e:
            self.error(e)

    def start(self):
        self.database = Database(self.database_path)
        self.database.start()
        atexit.register(self.end)
        if self.action == 'init':
            self.create_database()
        elif self.action == 'new':
            self.add_entry()
        elif self.action == 'list':
            self.show_list()
        else:
            self.error('action %s not recognized' % self.action, True)

    def end(self):
        self.database.end()
